SuperStrict

' Holzchopf - Halloween Horror Hunt
' BCC #89

Framework brl.glmax2d
Import brl.pngloader
Import brl.timer
Import brl.random

Import brl.basic

' consts
Const SCREEN_WIDTH:Int		= 1024
Const SCREEN_HEIGHT:Int		= 768
Const SCREEN_HERTZ:Int		= 50
Const SCREEN_FRAMETIME:Int	= 1000 / SCREEN_HERTZ

Const STATE_UNCHANGED:Int	= -1
Const STATE_MENU:Int		= 0
Const STATE_GAME:Int		= 1
Const STATE_GAMEOVER:Int	= 2

Const FIG_COUNT:Int				= 6
Const FIG_SIZE:Int				= 256
Const FIG_HALFSIZE:Int			= FIG_SIZE / 2
Global FIG_POSITIONS:Int[][]	= [[160, 160], [512, 160], [864, 160], [160, 448], [512, 448], [864, 448]]
Const FIG_FRAMES:Int			= 12
Const FIG_SAFETYDIST:Int		= 286
Const FIG_ROT_OMEGA:Float		= 4.0
Const FIG_ROT_AMP:Float			= 15
Const FIG_DIST_OFFSET:Float		= 1.1
Const FIG_DIST_OMEGA:Float		= 15.0
Const FIG_DIST_AMP:Float		= 0.05
Const FIG_ILLUMINATION_EXP:Int	= 2^14
Const THUMBNAIL_SIZE:Int		= 96

Const FLARE_SIZE:Int			= 256
Const FLARE_HALFSIZE:Int		= FLARE_SIZE / 2

Const HEALTH_MAX:Int		= 10000
Const HEALTH_VMIN:Int		= 10
Const HEALTH_VINC:Int		= 1
Const HEALTH_PENALTY:Int	= 2000

Const HIT_MISS:Int		= 0
Const HIT_RIGHT:Int		= 1
Const HIT_WRONG:Int		= 2

Const FIGURES_ACTIVE_MIN:Int		= 2
Const FIGURES_ACTIVE_INCTIME:Int	= 20000
Const FIGURES_CHANGETIME:Int		= 5000

' init app
AppTitle = "Halloween Horror Hunt"
Local timer:TTimer = CreateTimer( SCREEN_HERTZ )

SeedRnd(MilliSecs())
Graphics(SCREEN_WIDTH, SCREEN_HEIGHT)
HideMouse()

' load media
Global imgFigures:TImage = LoadAnimImage("gfx/figures.png", FIG_SIZE, FIG_SIZE, 0, FIG_FRAMES)
Global imgMasks:TImage = LoadAnimImage("gfx/figures.png", FIG_SIZE, FIG_SIZE, 0, FIG_FRAMES)
Global imgThumbnails:TImage = LoadAnimImage("gfx/thumbnails.png", THUMBNAIL_SIZE, THUMBNAIL_SIZE, 0, FIG_FRAMES)
Global imgFlare:TImage = LoadImage("gfx/flare.png")
Global imgTimeline:TImage = LoadImage("gfx/timeline.png")
Global imgPointer:TImage = LoadImage("gfx/pointer.png")

Global font:TFont = New TFont.Init("gfx/font")

MidHandleImage(imgFigures)
MidHandleImage(imgMasks)
MidHandleImage(imgThumbnails)
MidHandleImage(imgFlare)
MidHandleImage(imgPointer)

Global highscore:Int

LoadHighscore()

' global vars
Global mx:Int, my:Int, mh1:Int			' inputs
Global apptime:Int, gametime:Int		' timing
Global ticks:Int = 0					' timer ticks per loop

Global state:Int	= STATE_MENU		' current state
Global newstate:Int = STATE_UNCHANGED	' next state to change to

Global points:Int						' points, score, hits
Global clicks:Int						' mouse clicks

Global figurestock:Int[FIG_FRAMES]		' figures to choose from
Global figs:TFigure[FIG_COUNT]			' figure places on screen
Global figtofind:Int					' figure to find
Global figsactive:Int					' number of figures active (from stock)
Global figsonscreen:Int[FIG_FRAMES]		' which figures are visible
Global figchangetime:Int				' next time to change figures

Global health:Int						' current health
Global health_v:Int						' health decrease velocity

Rem
======================================================================
Main Loop
----------------------------------------------------------------------
End Rem
While Not AppTerminate()
	' state machine controller
	If newstate <> STATE_UNCHANGED then
		state = newstate
		newstate = STATE_UNCHANGED
	EndIf

	apptime :+ ticks * SCREEN_FRAMETIME

	' inputs
	mx = MouseX()
	my = MouseY()
	mh1 = MouseHit(MOUSE_LEFT)
	
	' computing and output
	Menu()
	Game()
	GameOver()

	' draw pointer
	SetColor(64,192,255)
	Local rot:Float = apptime /3.0
	SetRotation(-rot)
	DrawImage(imgPointer, mx+0.5, my+0.5)
	Plot(mx, my)

	ticks = WaitTimer(timer)
	Flip 0
Wend

Rem
======================================================================
Menu
----------------------------------------------------------------------
End Rem
Function Menu()
	If state <> STATE_MENU then Return
	
	SetClsColor(0, 0, 0)
	Cls
	
	SetBlend(ALPHABLEND)
	SetRotation(0)
	SetScale(1,1)
	SetColor(192,96,64)
	font.Draw("Highscore: " +highscore, 512, 192, ALIGN_CENTER, ALIGN_CENTER)
	
	' start game
	If DrawButton("Spiel starten", 384) then
		' rearrange figures in stock
		ShuffleInts(figurestock)
		figsactive = FIGURES_ACTIVE_MIN
		' place some figures on screen
		For Local i:Int = 0 Until FIG_COUNT
			figs[i] = New TFigure.Init(FIG_POSITIONS[i][0], FIG_POSITIONS[i][1])
			figs[i].SetFace( figurestock[Rand(0, figsactive -1)] )
		Next
		figtofind = PickRandomOnScreenFigure()
		' reset game vars
		points = 0
		clicks = 0
		health = HEALTH_MAX
		health_v = HEALTH_VMIN
		gametime = 0
		figchangetime = gametime + FIGURES_CHANGETIME
		newstate = STATE_GAME
	EndIf
	' quit game
	If DrawButton("Spiel beenden", 576) then
		End
	EndIf
End Function

Rem
======================================================================
Game
----------------------------------------------------------------------
End Rem
Function Game()
	If state <> STATE_GAME then Return

	SetClsColor(255,255,255)
	Cls

	gametime :+ ticks * SCREEN_FRAMETIME
	health :- health_v *ticks

	' increase number of possible figures with increasing time
	If figsactive < FIG_FRAMES then
		figsactive = FIGURES_ACTIVE_MIN +gametime / FIGURES_ACTIVE_INCTIME
	EndIf
	
	UpdateFigures()
	
	' try hitting something
	If mh1 then
		clicks :+ 1
		health_v :+ HEALTH_VINC
		' check if any figure was hit
		Local hit:Int = HIT_MISS
		For Local fig:TFigure = EachIn figs
			' actual hit
			If fig.hover = True then
				' hit right figure
				If fig.frame = figtofind then
					hit = HIT_RIGHT
					fig.frame = -1
				' hit wrong figure
				Else
					hit = HIT_WRONG
				EndIf
			EndIf
		Next
		' according to players aiming skill, ...
		If hit = HIT_RIGHT then
			points :+ 1
			health = HEALTH_MAX
			figtofind = PickRandomOnScreenFigure()
		ElseIf hit = HIT_WRONG then
			health :- HEALTH_PENALTY
		EndIf
	EndIf
	
	' end game if health is 0
	If health <= 0 then
		health = 0
		newstate = STATE_GAMEOVER
		SaveHighscore()
	EndIf
	
	' change / respawn invisible figure
	Local changefig:Int
	For Local i:Int = 0 Until FIG_COUNT
		If Not figs[i].isvisible then
			changefig = False
			' respawn figure
			If figs[i].frame = -1 then
				changefig = True
			' change figure
			ElseIf gametime >= figchangetime then
				' if figure is beeing looked for, prevent extinction
				If figs[i].frame = figtofind then
					If CountFiguresOfFace( figtofind ) > 1 then
						changefig = True
					EndIf
				Else
					changefig = True
				EndIf
			EndIf
			' execute change
			If changefig then
				figs[i].SetFace( figurestock[Rand(0, figsactive -1)] )
			EndIf
		EndIf
	Next
	If gametime >= figchangetime then
		figchangetime :+ FIGURES_CHANGETIME
	EndIf
	
	DrawFigures()
	DrawLight()
	DrawHUD()

	' draw health bar
	SetBlend(ALPHABLEND)
	SetColor(255,255,255)
	Local wh:Int
	wh = 960 *health / HEALTH_MAX
	DrawSubImageRect(imgTimeline, 32,704, wh,32, 0,0, wh,32)
End Function

Rem
======================================================================
Game over screen
----------------------------------------------------------------------
End Rem
Function GameOver()
	If state <> STATE_GAMEOVER then Return

	cls

	DrawFigures()
	DrawLight()
	DrawHUD()
	
	SetColor(192,96,64)
	font.Draw("Das Spiel ist um", 512, 304, ALIGN_CENTER, ALIGN_CENTER)
	
	If DrawButton("OK", 720) then
		newstate = STATE_MENU
	EndIf
End Function

Rem
======================================================================
Draw button, return true if button is pressed
----------------------------------------------------------------------
End Rem
Function DrawButton:Int(pLabel:String, pY:Int)
	Local hover:Int = False
	If my >= pY -32 And my < pY +32 then
		hover = True
	EndIf

	SetBlend(ALPHABLEND)
	SetRotation(0)
	SetScale(1,1)
	If hover then
		SetColor(255,255,255)
	Else
		SetColor(192,96,64)
	EndIf
	font.Draw(pLabel, 512, pY, ALIGN_CENTER, ALIGN_CENTER)
	Return hover And mh1
End Function

Rem
======================================================================
Update figures with collision check
----------------------------------------------------------------------
End Rem
Function UpdateFigures()
	For Local fig:TFigure = EachIn figs
		fig.Update()
	Next
End Function

Rem
======================================================================
Draw figures and shadow
----------------------------------------------------------------------
End Rem
Function DrawFigures()
	For Local fig:TFigure = EachIn figs
		fig.Update()
		fig.Draw()
	Next
End Function

Rem
======================================================================
Draw spotlight
----------------------------------------------------------------------
End Rem
Function DrawLight()
	SetRotation(0)
	SetBlend(SHADEBLEND)
	SetColor(255,255,255)
	DrawImage(imgFlare, mx, my)
	Local x1:Int, x2:Int, y1:Int, y2:Int
	x1 = mx -FLARE_HALFSIZE
	x2 = mx +FLARE_HALFSIZE
	y1 = my -FLARE_HALFSIZE
	y2 = my +FLARE_HALFSIZE
	SetColor(0,0,0)
	DrawRect(0,0, SCREEN_WIDTH,y1)
	DrawRect(0,y2, SCREEN_WIDTH,SCREEN_HEIGHT-y2)
	DrawRect(0,y1, x1,FLARE_SIZE)
	DrawRect(x2,y1, SCREEN_WIDTH-x2,FLARE_SIZE)
End Function

Rem
======================================================================
Draw HUD
----------------------------------------------------------------------
End Rem
Function DrawHUD()
	SetRotation(0)
	SetBlend(ALPHABLEND)
	SetColor(192,96,64)
	font.Draw("finde:", 32,640, , ALIGN_CENTER)
	font.Draw("Punkte: " +points, 992,640, ALIGN_RIGHT, ALIGN_CENTER)
	SetColor(255,255,255)
	DrawImage(imgThumbnails, 240,640, figtofind)
End Function

Rem
======================================================================
Shuffle pArray
Fill pArray with random numbers so every number from
0 to pArray.length -1 only occurs once
----------------------------------------------------------------------
End Rem
Function ShuffleInts(pArray:Int[])
	Local num_cnt:Int = pArray.length
	' empty array passed
	If num_cnt = 0 then
		Return
	' one-sized array passed
	ElseIf num_cnt = 1 then
		pArray[0] = 0
		Return
	EndIf
	Local numbers:Int[num_cnt]
	' fill number stock
	For Local i:Int = 0 Until num_cnt
		numbers[i] = i
	Next
	' fill array
	Local num_idx:Int
	For Local i:Int = 0 Until num_cnt
		' only one number left:
		If num_cnt = 1 then
			' pick that one number
			pArray[i] = numbers[0]
		' multiple numbers available:
		Else
			' pick random available number
			num_idx = Rand(0, num_cnt -1)
			pArray[i] = numbers[num_idx]
			' replace picked number with last number in stock
			numbers[num_idx] = numbers[num_cnt -1]
			' decrease count of remaining numbers
			num_cnt :- 1
		EndIf
	Next
End Function

Rem
======================================================================
Pick random figure (one which is actually on screen)
----------------------------------------------------------------------
End Rem
Function PickRandomOnScreenFigure:Int()
	' look, which figures are on screen
	For Local i:Int = 0 Until FIG_FRAMES
		figsonscreen[i] = 0
	Next
	For Local i:Int = 0 Until FIG_COUNT
		figsonscreen[ figs[i].frame ] = True
	Next
	' pick one of the figures on screen
	Local fig:Int
	Repeat
		fig = Rnd(FIG_FRAMES)
	Until figsonscreen[fig] = True
	Return fig
End Function

Rem
======================================================================
Count figures on screen of given face
----------------------------------------------------------------------
End Rem
Function CountFiguresOfFace:Int(pFace:Int)
	Local cnt:Int = 0
	For Local i:Int = 0 Until FIG_COUNT
		If figs[i].frame = pFace then
			cnt :+ 1
		EndIf
	Next
	Return cnt
End Function

Rem
======================================================================
Load highscore
----------------------------------------------------------------------
End Rem
Function LoadHighscore()
	Local file:TStream = ReadFile("highscore.dat")
	If file then
		highscore = file.ReadInt()
		file.Close()
	EndIf
End Function

Rem
======================================================================
Save highscore
----------------------------------------------------------------------
End Rem
Function SaveHighscore()
	If points <= highscore then Return
	Local file:TStream = WriteFile("highscore.dat")
	If file then
		file.WriteInt(points)
		file.Close()
	EndIf
	highscore = points
End Function

Rem
======================================================================
Figures
----------------------------------------------------------------------
End Rem
Type TFigure
	Field posx:Int, posy:Int
	Field frame:Int
	Field dx:Int, dy:Int
	Field rot:Float, dist:Float
	
	Field isvisible:Int
	Field hover:Int
	
	' initiate new figure
	Method Init:TFigure(pX:Int, pY:Int)
		posx = pX
		posy = pY
		Return Self
	End Method
	
	' set new figure face
	Method SetFace(pFrame:Int)
		frame = pFrame
	End Method

	' update, move and collision check
	Method Update()
		' reset hover flag
		hover = False
		
		rot = sin(gametime / FIG_ROT_OMEGA) * FIG_ROT_AMP
		dist = FIG_DIST_OFFSET +sin(gametime / FIG_DIST_OMEGA) * FIG_DIST_AMP
		dx = mx -posx
		dy = my -posy
		
		' figure visible?
		If Abs(dx) >= FIG_SAFETYDIST Or Abs(dy) >= FIG_SAFETYDIST then
			isvisible = False
		Else
			isvisible = True
		EndIf

		If frame = -1 then
			Return
		EndIf	
		
		' collision check
		Local projx:Int, projy:Int
		' transform mouse coordinates in figure frame
		projx = dx *cos(-rot) -dy *sin(-rot)
		projy = dx *sin(-rot) +dy *cos(-rot)
		' transformed mouse coordinates whithin figure bounding box
		If projx >= -FIG_HALFSIZE And projx < FIG_HALFSIZE And projy >= -FIG_HALFSIZE And projy < FIG_HALFSIZE then
			' pixel precise collision check
			Local pm:TPixmap = LockImage(imgFigures, frame, True, False)
			Local argb:Int = pm.ReadPixel(projx + FIG_HALFSIZE, projy + FIG_HALFSIZE)
			' collide if pixel is not absolutely transparent
			If (argb & $ff000000) <> $00 then
				hover = True
			EndIf
			UnlockImage(imgFigures, frame)
		EndIf
	End Method

	' draw figure with shadow
	Method Draw()
		If frame = -1 then
			Return
		EndIf
		' draw shadow with a slight offset
		Local off:Float
		off = dist /10.0
		SetRotation(rot)
		SetBlend(ALPHABLEND)
		SetColor(0,0,0)
		SetScale(dist, dist)
		DrawImage(imgMasks, posx -dx *off, posy -dy *off, frame)
		
		' draw figure, illuminated according to distance of light
		Local dd:Float, grey:Int
		dd = dx*dx +dy*dy
		grey = 255 * exp(-dd / FIG_ILLUMINATION_EXP)
		SetColor(grey, grey, grey)
		SetScale(1,1)
		DrawImage(imgFigures, posx, posy, frame)
	End Method
End Type


Rem
======================================================================
Simple bitmap font type
----------------------------------------------------------------------
End Rem
Const ALIGN_LEFT:Int	= 0
Const ALIGN_CENTER:Int	= 1
Const ALIGN_RIGHT:Int	= 2
Const ALIGN_TOP:Int		= 0
Const ALIGN_BASELINE:Int= 2
' image font, simplified
Type TFont
	Field image:TImage, cellsize:Int
	Field offset_x:Int, offset_y:Int, spacing:Int, fontheight:Int
	Field widths:Int[] = New Int[256]
	
	' loads a font, pName.png and pName.dat must exist!
	Method Init:TFont(pName:String)
		' first: read font info (.dat) file, containing char widths, cell size and x/y offset
		Local file:TStream = ReadFile(pName +".dat")
		Local char:Byte, buffer:String, value:Int
		Local eot:Byte=False ' used to indicate end of text (white space or file ending)
		Local idx:Int = 0
		' read as long as there is data in file
		While Not file.Eof()
			char = file.ReadByte()
			' numerical char 48 == ascii:0, 57 == ascii:9
			If char >= 48 And char <= 57 then
				' add char to buffer
				buffer :+ Chr(char)
				eot = False
			Else
				eot = True
			EndIf
			' set end of text at end of file
			If file.Eof() then eot=True
			' any other char
			If eot = True And buffer <> "" then
				' read integer value from buffer
				value = Int(buffer)
				' depending on idx, interpret
				Rem
					0...255		width
					256			cellsize
					257,258		x/y-offset
					259			spacing
					260			font height
				end Rem
				If idx <= 255 then
					widths[idx] = value
				ElseIf idx = 256 then
					cellsize = value
				ElseIf idx = 257 then
					offset_x = value
				ElseIf idx = 258 then
					offset_y = value
				ElseIf idx = 259 then
					spacing = value
				ElseIf idx = 260 then
					fontheight = value
				EndIf
				' get ready for next value
				buffer = ""
				idx :+ 1
			EndIf
		Wend
		file.Close()
		' second: load image file
		image = LoadAnimImage(pName +".png", cellsize, cellsize, 0, 256)
		Return self
	End Method
	
	' draw a text
	Method Draw(pText:String, pX:Float, pY:Float, pHorizontalAlign:Int=0, pVerticalAlign:Int=ALIGN_BASELINE)
		'SetBlend(SOLIDBLEND)
		' ^ due to possible scaling, use float for position
		' save active scaling
		Local scale_x:Float, scale_y:Float
		GetScale(scale_x, scale_y)
		' remove offset_x (aka the white space left of each char)
		pX :- offset_x *scale_x
		' position according to horizontal aligning
		If pHorizontalAlign = ALIGN_CENTER
			pX :- GetWidth(pText) *0.5 *scale_x
		ElseIf pHorizontalAlign = ALIGN_RIGHT
			pX :- GetWidth(pText) *scale_x
		EndIf
		' remove offset_y
		pY :- offset_y *scale_y
		' position according to vertical aligning
		If pVerticalAlign = ALIGN_CENTER
			pY :+ GetHeight(pText) *0.5 *scale_y
		ElseIf pVerticalAlign = ALIGN_TOP
			pY :+ GetHeight(pText) *scale_y
		EndIf
		' draw text char by char
		Local char:Byte
		For Local i:Int = 0 Until pText.length
			char = pText[i]
			' draw only valid chars
			If char <= 255 then
				DrawImage(image, pX, pY, char)
				' move drawing position
				pX :+ (widths[char] -offset_x +spacing) *scale_x
			EndIf
		Next
	End Method
	
	' return width of given text
	Method GetWidth:Float(pText:String)
		Local width:Float
		' sum up widths char by char
		Local char:Byte
		For Local i:Int = 0 Until pText.length
			char = pText[i]
			If char <= 255 then
				width :+ (widths[char] -offset_x +spacing)
			EndIf
		Next
		' subtract spacing once, for last char of text
		width :- spacing
		Return width
	End Method
	
	' return height of given text
	' function only exists for consistency
	Method GetHeight:Float(pText:String)
		Return fontheight
	End Method
End Type

